"""Parsers."""

import re
import struct

NODE_LOOKUP = {
    '10': 'master',
    '11': 'echo',
    '20': 'gateway',
}

ATTR_LOOKUP = {
    'cmd': 'command',
    'bat': 'battery',
    'temp': 'temperature',
    'hum': 'humidity',
    'rev': 'revision',
    'ver': 'version',
}

VALUE_TRANSLATION = {
    # 'temp': lambda t: int(t, 16)/10,
    'temp': lambda t: struct.unpack('>h', bytes.fromhex(t))[0]/10,
    'hum': int,
}

BANNER_RE = (r'(?P<hardware>[a-zA-Z\s]+) - (?P<firmware>[a-zA-Z\s]+) '
            r'V(?P<version>[0-9\.]+) - R(?P<revision>[0-9\.]+)')


def parse_packet(packet):
    """Break packet down into primitives, and do basic interpretation.

    >>> parse_packet('20;06;Kaku;ID=41;SWITCH=1;CMD=ON;') == {
    ...     'node': 'gateway',
    ...     'protocol': 'kaku',
    ...     'id': '000041',
    ...     'switch': '1',
    ...     'command': 'on',
    ... }
    True
    """
    node_id, _, protocol, attrs = packet.split(';', 3)

    data = {
        'node': NODE_LOOKUP[node_id],
    }

    # make exception for version response
    if '=' in protocol:
        attrs = protocol + ';' + attrs
        protocol = 'version'

    # no attributes but instead the welcome banner
    elif 'RFLink Gateway' in protocol:
        data.update(parse_banner(protocol))
        protocol = 'banner'

    elif protocol == 'PONG':
        data['ping'] = protocol.lower()

    # failure response
    elif protocol == 'CMD UNKNOWN':
        data['response'] = 'command unknown'
        data['ok'] = False

    # ok response
    elif protocol == 'OK':
        data['ok'] = True

    # its a regular packet
    else:
        data['protocol'] = protocol.lower()

    # convert key=value pairs where needed
    for attr in filter(None, attrs.strip(';').split(';')):
        key, value = attr.lower().split('=')
        if key in VALUE_TRANSLATION:
            value = VALUE_TRANSLATION.get(key)(value)
        key = ATTR_LOOKUP.get(key, key)
        data[key] = value

    # correct KaKu device address
    if data.get('protocol', '') == 'kaku' and len(data['id']) != 6:
        data['id'] = '0000' + data['id']

    return data


def parse_banner(banner):
    """Extract hardware/firmware name and version from banner."""
    return re.match(BANNER_RE, banner).groupdict()



"""Asyncio protocol implementation of RFlink."""
import asyncio
import logging
import re
from functools import partial

from serial_asyncio import create_serial_connection

from .parser import parse_packet

log = logging.getLogger(__name__)


class ProtocolBase(asyncio.Protocol):
    """Manage low level rflink protocol."""

    transport = None

    def __init__(self, loop, packet_callback=None):
        """Initialize class."""
        self.loop = loop
        self.packet_callback = packet_callback
        self.packet = ''
        self.start_packet = re.compile('^(10|11|20);').match
        self.buffer = ''

    def connection_made(self, transport):
        """Just logging for now."""
        self.transport = transport
        log.debug('connected')

    def data_received(self, data):
        """Add incoming data to buffer."""
        data = data.decode()
        log.debug('received data: %s', data.strip())
        self.buffer += data
        self.handle_lines()

    def handle_lines(self):
        """Assemble incoming data into per-line packets."""
        while "\r\n" in self.buffer:
            line, self.buffer = self.buffer.split("\r\n", 1)
            if self.start_packet(line):
                self.handle_raw_packet(line)

    def connection_lost(self, exc):
        """Stop when connection is lost."""
        log.error('disconnected')
        self.loop.stop()


class PacketHandling:
    """Handle translating rflink packets."""

    def handle_raw_packet(self, raw_packet):
        """Callback for handling incoming raw packets."""
        log.debug('got packet: %s', raw_packet)
        packet = None
        try:
            packet = parse_packet(raw_packet)
        except:
            log.exception('failed to parse packet: %s', packet)

        if packet:
            self.handle_packet(packet)

    def handle_packet(self, packet):
        """Callback for handling incoming parsed packets."""
        log.debug('parsed packet: %s', packet)
        if self.packet_callback:
            self.packet_callback(packet)
        else:
            print(packet)

    def send_command(self, protocol, address, switch, action):
        """Send device command to rflink gateway."""
        command = [protocol, address, switch, action]
        log.debug('sending command: %s', command)
        self.send_packet(command)

    def send_packet(self, fields):
        """Concat fields and send packet to gateway."""
        fields = ['10'] + fields
        data = ';'.join(fields + ['\r\n']).encode()
        self.transport.write(data)


class RflinkProtocol(PacketHandling, ProtocolBase):
    """Combine low and high level protocol handling."""


class InverterProtocol(RflinkProtocol):
    """Invert switch commands received and send them out."""

    def handle_packet(self, packet):
        """Handle incoming packet from rflink gateway."""
        if packet.get('switch'):
            if packet['command'] == 'on':
                cmd = 'off'
            else:
                cmd = 'on'
            self.send_command(packet['protocol'], packet['id'],
                              packet['switch'], cmd)


class RepeaterProtocol(RflinkProtocol):
    """Repeat switch commands received."""

    def handle_packet(self, packet):
        """Handle incoming packet from rflink gateway."""
        if packet.get('switch'):
            self.send_command(packet['protocol'], packet['id'],
                              packet['switch'], packet['command'])


def create_rflink_connection(*args, **kwargs):
    """Create Rflink manager class, returns transport coroutine."""
    loop = kwargs.pop('loop', asyncio.get_event_loop())

    # use default protocol if not specified
    rflink_protocol = kwargs.pop('protocol', RflinkProtocol)
    packet_callback = kwargs.pop('packet_callback', None)
    protocol = partial(rflink_protocol, loop, packet_callback=packet_callback)

    # setup serial connection if no transport specified
    host = kwargs.pop('host', None)
    port = kwargs.pop('port')
    if host:
        conn = loop.create_connection(protocol, host, port)
    else:
        baud = kwargs.get('baud', 57600)
        conn = create_serial_connection(loop, protocol, port, baud)

    return conn



"""Command line interface for rflink library.

Usage:
  rflink [options]
  rflink (-h | --help)
  rflink --version

Options:
  -p --port=<port>  Serial port to connect to [default: /dev/ttyACM0],
                      or TCP port in TCP mode.
  --baud=<baud>     Serial baud rate [default: 57600].
  --host=<host>     TCP mode, connect to host instead of serial port.
  -m=<handling>     How to handle incoming packets [default: print].
  -h --help         Show this screen.
  -v --verbose      Increase verbosity
  --version         Show version.

"""

import asyncio
import logging
import sys

import pkg_resources
from docopt import docopt

from .protocol import (
    InverterProtocol,
    RepeaterProtocol,
    RflinkProtocol,
    create_rflink_connection
)

PROTOCOLS = {
    'print': RflinkProtocol,
    'invert': InverterProtocol,
    'repeat': RepeaterProtocol,
}


def main(argv=sys.argv[1:], loop=None):
    """Parse argument and setup main program loop."""
    args = docopt(__doc__, argv=argv, version=pkg_resources.require('rflink')[0].version)

    if args['--verbose']:
        level = logging.DEBUG
    else:
        level = logging.INFO
    logging.basicConfig(level=level)

    if not loop:
        loop = asyncio.get_event_loop()

    protocol = PROTOCOLS[args['-m']]
    conn = create_rflink_connection(
        protocol=protocol,
        host=args['--host'],
        port=args['--port'],
        baud=args['--baud'],
        loop=loop,
    )

    transport, protocol = loop.run_until_complete(conn)

    try:
        loop.run_forever()
    except KeyboardInterrupt:
        # cleanup connection
        transport.close()
        loop.run_forever()
    finally:
        loop.close()



"""Library and CLI tools for interacting with RFlink 433MHz transceiver."""

from codecs import open
from os import path

from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='rflink',

    version='0.0.5',

    description=__doc__,
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/aequitas/python-rflink',

    # Author details
    author='Johan Bloemberg',
    author_email='github@ijohan.nl',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='rflink 433mhz domotica',

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    install_requires=[
        'docopt',
        'pyserial-asyncio',
    ],

    # # List additional groups of dependencies here (e.g. development
    # # dependencies). You can install these using the following syntax,
    # # for example:
    # # $ pip install -e .[dev,test]
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'test': ['coverage'],
    # },

    # # If there are data files included in your packages that need to be
    # # installed, specify them here.  If using Python 2.6 or less, then these
    # # have to be included in MANIFEST.in as well.
    # package_data={
    #     'sample': ['package_data.dat'],
    # },

    # # Although 'package_data' is the preferred approach, in some case you may
    # # need to place data files outside of your packages. See:
    # # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    # data_files=[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'rflink=rflink.__main__:main',
        ],
    },
)
